@@ -1,10 +1,9 @@
 from __future__ import annotations
 
 import asyncio
-import contextlib
 import logging
 import struct
-from collections.abc import Callable
+from collections.abc import Callable, Iterable
 from dataclasses import replace
 from typing import Any, TypeVar, cast
 
@@ -38,23 +37,25 @@ WrapFuncType = TypeVar("WrapFuncType", bound=Callable[..., Any])
 
 DEFAULT_ATTEMPTS = 4
 
+DISCONNECT_DELAY = 5.1
+
 # How long to wait before processing an advertisement change
-ADV_UPDATE_COALESCE_SECONDS = 6.99
+ADV_UPDATE_COALESCE_SECONDS = 0.05
 
 # How long to wait before processing the first update
-FIRST_UPDATE_COALESCE_SECONDS = 0.50
+FIRST_UPDATE_COALESCE_SECONDS = 0.01
 
 # How long to wait before processing a HomeKit advertisement change
-HK_UPDATE_COALESCE_SECONDS = 2.00
+HK_UPDATE_COALESCE_SECONDS = 0.025
 
 # How long to wait before processing a manual update request
-MANUAL_UPDATE_COALESCE_SECONDS = 0.75
+MANUAL_UPDATE_COALESCE_SECONDS = 0.05
 
 # How long to wait to query the lock after an operation to make sure its not jammed
 POST_OPERATION_SYNC_TIME = 10.00
 
 # How long to wait if we get an update storm from the lock
-UPDATE_IN_PROGRESS_DEFER_SECONDS = 29.50
+UPDATE_IN_PROGRESS_DEFER_SECONDS = 1.0
 
 RETRY_BACKOFF_EXCEPTIONS = (BleakDBusError, DisconnectedError)
 
@@ -65,21 +66,6 @@ RETRY_EXCEPTIONS = (ResponseError, *BLEAK_RETRY_EXCEPTIONS)
 VALID_ADV_VALUES = {0, 1}
 
 
-def cancelable_operation(func: WrapFuncType) -> WrapFuncType:
-    """Define a wrapper to make mutually exclusive operations cancelable."""
-
-    async def _async_wrap_cancelable_operation(
-        self: "PushLock", *args: Any, **kwargs: Any
-    ) -> None:
-        await self._cancel_in_progress_operation()
-        await self._cancel_in_progress_update()
-        self._operation_task = asyncio.create_task(func(self, *args, **kwargs))
-        await self._operation_task
-        self._operation_task = None
-
-    return cast(WrapFuncType, _async_wrap_cancelable_operation)
-
-
 def operation_lock(func: WrapFuncType) -> WrapFuncType:
     """Define a wrapper to only allow a single operation at a time."""
 
@@ -179,6 +165,7 @@ class PushLock:
         key: str | None = None,
         key_index: int | None = None,
         advertisement_data: AdvertisementData | None = None,
+        idle_disconnect_delay: float = DISCONNECT_DELAY,
     ) -> None:
         """Init the lock watcher."""
         if local_name is None and address is None:
@@ -204,13 +191,16 @@ class PushLock:
         self._callbacks: list[
             Callable[[LockState, LockInfo, ConnectionInfo], None]
         ] = []
-        self._operation_task: asyncio.Task | None = None  # type: ignore[type-arg]
-        self._update_task: asyncio.Task | None = None  # type: ignore[type-arg]
         self._debounce_lock = asyncio.Lock()
         self.loop = asyncio._get_running_loop()
         self._cancel_deferred_update: asyncio.TimerHandle | None = None
+        self._cancel_post_lock_op_sync: asyncio.TimerHandle | None = None
         self.last_error: str | None = None
         self.auth_error = False
+        self._client: Lock | None = None
+        self._connect_lock = asyncio.Lock()
+        self._disconnect_timer: asyncio.TimerHandle | None = None
+        self._idle_disconnect_delay = idle_disconnect_delay
 
     @property
     def local_name(self) -> str | None:
@@ -312,87 +302,176 @@ class PushLock:
             self._lock_key,
             self._lock_key_index,
             self.name,
+            self._state_callback,
             self._lock_info,
         )
 
-    async def _cancel_any_update(self) -> None:
-        """Cancel any update task."""
-        await asyncio.sleep(0)
-        await self._cancel_in_progress_update()
+    def _reset_disconnect_timer(self) -> None:
+        """Reset disconnect timer."""
+        self._cancel_disconnect_timer()
+        self._expected_disconnect = False
+        self._disconnect_timer = self.loop.call_later(
+            self._idle_disconnect_delay, self._disconnect_with_timer
+        )
 
-    async def _cancel_in_progress_update(self) -> None:
-        """Cancel any in progress update task."""
-        if self._update_task:
-            _LOGGER.debug("Canceling in progress update: %s", self._update_task)
-            self._update_task.cancel()
-            with contextlib.suppress(asyncio.CancelledError, Exception):
-                await self._update_task
-            self._update_task = None
+    async def _execute_forced_disconnect(self) -> None:
+        """Execute forced disconnection."""
+        self._cancel_disconnect_timer()
+        _LOGGER.debug(
+            "%s: Executing forced disconnect",
+            self.name,
+        )
+        await self._execute_disconnect()
 
-    async def _cancel_in_progress_operation(self) -> None:
-        """Cancel any in progress operation task."""
-        if self._operation_task:
-            _LOGGER.debug("Canceling in progress task: %s", self._operation_task)
-            self._operation_task.cancel()
-            with contextlib.suppress(asyncio.CancelledError, Exception):
-                await self._operation_task
-            self._operation_task = None
+    def _disconnect_with_timer(self) -> None:
+        """Disconnect from device."""
+        if self._operation_lock.locked():
+            self._reset_disconnect_timer()
+            return
+        self._cancel_disconnect_timer()
+        asyncio.create_task(self._execute_timed_disconnect())
+
+    def _cancel_disconnect_timer(self) -> None:
+        """Cancel disconnect timer."""
+        if self._disconnect_timer:
+            self._disconnect_timer.cancel()
+            self._disconnect_timer = None
+
+    async def _execute_timed_disconnect(self) -> None:
+        """Execute timed disconnection."""
+        _LOGGER.debug(
+            "%s: Executing timed disconnect after timeout of %s",
+            self.name,
+            DISCONNECT_DELAY,
+        )
+        await self._execute_disconnect()
+
+    async def _execute_disconnect(self) -> None:
+        """Execute disconnection."""
+        async with self._connect_lock:
+            if self._disconnect_timer:  # If the timer was reset, don't disconnect
+                return
+            client = self._client
+            self._client = None
+            if client and client.is_connected:
+                _LOGGER.debug("%s: Disconnecting", self.name)
+                await client.disconnect()
+                _LOGGER.debug("%s: Disconnect completed", self.name)
+
+    async def _ensure_connected(self) -> Lock:
+        """Ensure connection to device is established."""
+        if self._connect_lock.locked():
+            self._reset_disconnect_timer()
+            _LOGGER.debug(
+                "%s: Connection already in progress, waiting for it to complete",
+                self.name,
+            )
+        if self._client and self._client.is_connected:
+            self._reset_disconnect_timer()
+            return self._client
+        async with self._connect_lock:
+            # Check again while holding the lock
+            if self._client and self._client.is_connected:
+                self._reset_disconnect_timer()
+                return self._client
+            self._client = self._get_lock_instance()
+            await self._client.connect()
+            self._reset_disconnect_timer()
+            return self._client
 
-    @cancelable_operation
-    @operation_lock
-    @retry_bluetooth_connection_error
     async def lock(self) -> None:
         """Lock the lock."""
-        _LOGGER.debug("%s: Starting lock", self.name)
-        await self._cancel_any_update()
-        self._callback_state(
-            LockState(LockStatus.LOCKING, self.door_status, self._battery_state)
-        )
-        lock = self._get_lock_instance()
-        try:
-            async with lock:
-                await lock.force_lock()
-        except Exception:
-            self._callback_state(
-                LockState(LockStatus.UNKNOWN, self.door_status, self._battery_state)
-            )
-            raise
-        self._callback_state(
-            LockState(LockStatus.LOCKED, self.door_status, self._battery_state)
-        )
-        await self._cancel_any_update()
-        self._schedule_update(POST_OPERATION_SYNC_TIME)
-        _LOGGER.debug("%s: Finished lock", self.name)
+        self._update_any_state([LockStatus.LOCKING])
+        self._cancel_update()
+        await self._lock_with_op_lock()
 
-    @cancelable_operation
     @operation_lock
     @retry_bluetooth_connection_error
+    async def _lock_with_op_lock(self) -> None:
+        """Lock the lock."""
+        _LOGGER.debug("%s: Starting lock", self.name)
+        self._update_any_state([LockStatus.LOCKING])
+        self._cancel_update()
+        try:
+            lock = await self._ensure_connected()
+            self._cancel_update()
+            await lock.force_lock()
+        except Exception:
+            self._update_any_state([LockStatus.UNKNOWN])
+            raise
+        self._update_any_state([LockStatus.LOCKED])
+        _LOGGER.debug("%s: Finished lock", self.name)
+
     async def unlock(self) -> None:
+        """Unlock the lock."""
+        self._update_any_state([LockStatus.UNLOCKING])
+        self._cancel_update()
+        await self._unlock_with_op_lock()
+
+    @operation_lock
+    @retry_bluetooth_connection_error
+    async def _unlock_with_op_lock(self) -> None:
         """Unlock the lock."""
         _LOGGER.debug("%s: Starting unlock", self.name)
-        await self._cancel_any_update()
-        self._callback_state(
-            LockState(LockStatus.UNLOCKING, self.door_status, self._battery_state)
-        )
-        lock = self._get_lock_instance()
+        self._update_any_state([LockStatus.UNLOCKING])
+        self._cancel_update()
         try:
-            async with lock:
-                await lock.force_unlock()
+            lock = await self._ensure_connected()
+            self._cancel_update()
+            await lock.force_unlock()
         except Exception:
-            self._callback_state(
-                LockState(LockStatus.UNKNOWN, self.door_status, self._battery_state)
-            )
+            self._update_any_state([LockStatus.UNKNOWN])
             raise
-        self._callback_state(
-            LockState(LockStatus.UNLOCKED, self.door_status, self._battery_state)
-        )
-        await self._cancel_any_update()
-        self._schedule_update(POST_OPERATION_SYNC_TIME)
+        self._update_any_state([LockStatus.UNLOCKED])
         _LOGGER.debug("%s: Finished unlock", self.name)
 
+    def _state_callback(
+        self, states: Iterable[LockStatus | DoorStatus | BatteryState]
+    ) -> None:
+        """Handle state change."""
+        self._reset_disconnect_timer()
+        self._update_any_state(states)
+
+    def _update_any_state(
+        self, states: Iterable[LockStatus | DoorStatus | BatteryState]
+    ) -> None:
+        _LOGGER.debug("%s: State changed: %s", self.name, states)
+        lock_state = self._lock_state or LockState(
+            self.lock_status, self.door_status, self.battery
+        )
+        original_lock_status = lock_state.lock
+        for state in states:
+            if isinstance(state, LockStatus):
+                lock_state = replace(lock_state, lock=state)
+            elif isinstance(state, DoorStatus):
+                lock_state = replace(lock_state, door=state)
+            elif isinstance(state, BatteryState):
+                if state.voltage <= 3.0:
+                    _LOGGER.debug(
+                        "%s: Battery voltage is impossible: %s",
+                        self.name,
+                        state.voltage,
+                    )
+                    continue
+                lock_state = replace(lock_state, battery=state)
+            else:
+                raise ValueError(f"Unexpected state type: {state}")
+
+        if (
+            original_lock_status != lock_state.lock
+            and original_lock_status != LockStatus.UNKNOWN
+        ):
+            self._schedule_resync()
+
+        self._callback_state(lock_state)
+
     async def update(self) -> None:
         """Request that status be updated."""
-        self._schedule_update(MANUAL_UPDATE_COALESCE_SECONDS)
+        self._schedule_update(
+            0
+            if self._client and self._client.is_connected
+            else MANUAL_UPDATE_COALESCE_SECONDS
+        )
 
     async def validate(self) -> None:
         """Validate lock credentials."""
@@ -404,15 +483,18 @@ class PushLock:
     @retry_bluetooth_connection_error
     async def _update(self) -> LockState:
         """Update the lock state."""
-        _LOGGER.debug("%s: Starting update", self.name)
-        lock = self._get_lock_instance()
+        has_lock_info = self._lock_info is not None
+
+        _LOGGER.debug(
+            "%s: Starting update (has_lock_info: %s)", self.name, has_lock_info
+        )
         try:
-            async with lock:
-                if not self._lock_info:
-                    self._lock_info = await lock.lock_info()
-                state = await lock.status()
-                self._battery_state = await lock.battery()
-                state = replace(state, battery=self._battery_state)
+            lock = await self._ensure_connected()
+            if not self._lock_info:
+                self._lock_info = await lock.lock_info()
+            state = await lock.status()
+            self._battery_state = await lock.battery()
+            state = replace(state, battery=self._battery_state)
         except asyncio.CancelledError:
             _LOGGER.debug(
                 "%s: In-progress update canceled due "
@@ -422,6 +504,14 @@ class PushLock:
             raise
         _LOGGER.debug("%s: Finished update", self.name)
         self._callback_state(state)
+
+        if not has_lock_info:
+            # On first update free up the connection
+            # so we can bring other locks online if
+            # the bluetooth adapter is out of connections
+            # slots
+            await self._execute_forced_disconnect()
+
         return state
 
     def _callback_state(self, lock_state: LockState) -> None:
@@ -481,28 +571,28 @@ class PushLock:
                 #
                 # if len(mfr_data[APPLE_MFR_ID]) > 20 and YALE_MFR_ID not in mfr_data:
                 # mfr_data[YALE_MFR_ID] = mfr_data[APPLE_MFR_ID][20:]
-                if hk_state != self._last_hk_state:
-                    if self._last_hk_state == -1:
-                        next_update = FIRST_UPDATE_COALESCE_SECONDS
-                    else:
-                        next_update = HK_UPDATE_COALESCE_SECONDS
-                    self._last_hk_state = hk_state
+                if self._last_hk_state == -1:
+                    # We haven't seen a HomeKit state yet so we schedule an update
+                    next_update = FIRST_UPDATE_COALESCE_SECONDS
+                elif hk_state != self._last_hk_state:
+                    next_update = HK_UPDATE_COALESCE_SECONDS
+                self._last_hk_state = hk_state
             elif first_byte == HAP_ENCRYPTED_FIRST_BYTE:
                 # Encrypted data, we don't know how to decrypt it
                 # but we know its a state change so we schedule an update
                 next_update = HK_UPDATE_COALESCE_SECONDS
         if YALE_MFR_ID in mfr_data:
             current_value = mfr_data[YALE_MFR_ID][0]
-            if (
-                current_value in VALID_ADV_VALUES
-                and current_value != self._last_adv_value
-            ):
-                if not next_update:
-                    if self._last_adv_value == -1:
-                        next_update = FIRST_UPDATE_COALESCE_SECONDS
-                    else:
-                        next_update = ADV_UPDATE_COALESCE_SECONDS
-                self._last_adv_value = current_value
+            if not next_update:
+                if self._last_adv_value == -1:
+                    # We haven't seen a valid value yet so we schedule an update
+                    next_update = FIRST_UPDATE_COALESCE_SECONDS
+                elif (
+                    current_value in VALID_ADV_VALUES
+                    and current_value != self._last_adv_value
+                ):
+                    next_update = ADV_UPDATE_COALESCE_SECONDS
+            self._last_adv_value = current_value
         if adv_debug_enabled:
             scheduled_update = None
             if self._cancel_deferred_update:
@@ -535,10 +625,34 @@ class PushLock:
 
         def _cancel() -> None:
             self._running = False
-            asyncio.create_task(self._cancel_in_progress_update())
+            self._cancel_resync()
+            asyncio.create_task(self._execute_forced_disconnect())
 
         return _cancel
 
+    def _cancel_resync(self) -> None:
+        """Cancel a resync."""
+        if self._cancel_post_lock_op_sync:
+            self._cancel_post_lock_op_sync.cancel()
+            self._cancel_post_lock_op_sync = None
+
+    def _cancel_update(self) -> None:
+        """Cancel an update."""
+        if self._cancel_deferred_update:
+            self._cancel_deferred_update.cancel()
+            self._cancel_deferred_update = None
+
+    def _resync(self) -> None:
+        """Resync the lock state."""
+        self._schedule_update(0.01)
+
+    def _schedule_resync(self) -> None:
+        """Schedule a resync."""
+        self._cancel_resync()
+        self._cancel_post_lock_op_sync = self.loop.call_later(
+            POST_OPERATION_SYNC_TIME, self._resync
+        )
+
     def _schedule_update(self, seconds: float) -> None:
         """Schedule an update."""
         now = self.loop.time()
@@ -562,8 +676,7 @@ class PushLock:
                 )
                 return
             _LOGGER.debug("%s: Rescheduling update", self.name)
-            self._cancel_deferred_update.cancel()
-            self._cancel_deferred_update = None
+            self._cancel_update()
         _LOGGER.debug(
             "%s: Scheduling update to happen in %s seconds",
             self.name,
@@ -597,9 +710,7 @@ class PushLock:
             _LOGGER.debug("%s: Starting update", self.name)
             try:
                 self.last_error = "Could not connect"
-                self._update_task = asyncio.create_task(self._update())
-                await self._update_task
-                self._update_task = None
+                await self._update()
             except AuthError as ex:
                 self.auth_error = True
                 self.last_error = (
